/* (c) 1994 by Axel Zinser aka Fifi (fifi@hiss.han.de) */
#include	<stdio.h>
#include	<time.h>
#include	<string.h>
#include	<errno.h>
#include	<sys/types.h>
#include	<sys/fcntl.h>
#include	<sys/file.h>
#include	<sys/stat.h>
#include	<sys/param.h>
#include	<sys/ioctl.h>
#if defined(SCO)
#include	<sys/ndir.h>
#include	<sys/unistd.h>
#else
#include	<sys/dir.h>
#endif

#include	<sys/socket.h>
#include	<netinet/in.h>
#include	<netdb.h>

#include	"alloc.h"
#include	"version.h"
#include	"xstring.h"
#include	"config.h"
#include	"depress.h"

#if defined(sun4) || defined(__NeXT__) || defined(BSD)
# define STATFS
# include	<sys/vfs.h>
#endif
#if defined(SYSV) || defined(SVR3)
# define USTAT
# include	<ustat.h>
#endif

#define	BUFLEN	4096
char	buffer[BUFLEN];

typedef struct X {
	struct X	*next;
	char		*rcpt;
} charl;

long	mode = 0;
#define	CHECKSIZE	0x0001		/* check spool space */
#define	RDSTDIN		0x0002		/* read data stream without commu */
#define	RDSTDINC	0x0004		/* data stream may be compressed */
#define	WRBSMTP		0x0008		/* write smtp cmds */
#define	WRSPOOL		0x0010		/* write to spoolfile */
#define	WRPIPE		0x0020		/* write to pipe(mailer) */
#define	STATE_BSMTPD	(CHECKSIZE|WRBSMTP|WRSPOOL)
#define	STATE_RBSMTP	(RDSTDIN|RDSTDINC|WRPIPE)
#define	STATE_RSMTP	(CHECKSIZE|WRBSMTP|WRSPOOL)

char	vstring[50];
char	*localhost;
char	*actname = 0, *remotehost = NULL;
char	*sender = NULL;
charl	*recipient = NULL;
FILE	*lfp = NULL;
long	sysspace = 0;
/* FILE	*ofp = stdout; */
FILE	*ifp = stdin;
char	*prog = 0;

extern int 	rcpt(), rset(), mail(), noop(), vrfy(), helo(), quit(),
	        help(), ehlo(), debug(), bsmtp_send(), turn(), data(),
	        saml(), soml(), nimp();

struct FTab {
	char	*cmd;
	int	(*func)();
} ftab[] = {
	{ "HELO",	helo },
	{ "MAIL",	mail },
	{ "RCPT",	rcpt },
	{ "DATA",	data },
	{ "RSET",	rset },
	{ "NOOP",	noop },
	{ "QUIT",	quit },
	{ "HELP",	help },
	{ "VRFY",	vrfy },
	{ "EXPN",	vrfy },
#ifdef DEBUG
	{ "DEBUG",	debug },
#endif
#ifdef FULLCMD
	{ "SEND",	bsmtp_send },
	{ "SOML",	soml },
	{ "SAML",	saml },
	{ "TURN",	turn },
#endif
# if defined(SMTPEXT) || defined(RFC1652) || defined(RFC1653) /* { */
	{ "EHLO",	ehlo },
# endif /* } SMTPEXT | RFC1652 | RFC1653 */
	{ 0,		0 }
};

int	state = 0;
#define	HELO	0x001
#define	EHLO	0x008
#define	MAIL	0x002
#define	RCPT	0x004
#define	BIT7	0x010
#define BIT8	0x20

nimp() { fprintf(stdout,"502 Command not implemented\n"); }
bsmtp_send() { fprintf(stdout,"502 Command not implemented\n"); }
saml() { fprintf(stdout,"502 Command not implemented\n"); }
soml() { fprintf(stdout,"502 Command not implemented\n"); }
turn() { fprintf(stdout,"502 Command not implemented\n"); }

lock()
{
	int	o;

	while(!access(LOCKFILE,F_OK) || (o=open(LOCKFILE,O_WRONLY|O_CREAT|O_EXCL|O_NDELAY|O_SYNC,0644)) < 0)
		sleep(1);
# ifdef DEBUG
	fprintf(stderr,"*** %5d lock \"%s\"\n",getpid(),LOCKFILE);
	fflush(stderr);
# endif
	sprintf(buffer,"%d\n",getpid());
	write(o,buffer,strlen(buffer)+1);
	close(o);
#ifdef LOGFILE /* { */
	lfp = fopen(LOGFILE,"a+");
# ifdef DEBUG /* { */
	fprintf(stderr,"*** %5d open log \"%s\" %s\n",getpid(),
		LOGFILE,(lfp ? "suceeded" : "failed"));
	fflush(stderr);
# endif /* } DEBUG */
#endif /* } LOGFILE */
	return(0);
}

unlock()
{
#ifdef DEBUG /* { */
	fprintf(stderr,"*** %5d unlock \"%s\"\n",getpid(),LOCKFILE);
	if (access(LOCKFILE,F_OK)) fprintf(stderr,"*** %5d missing lockfile\n",getpid());
	fflush(stderr);
#endif /* } DEBUG */
#ifdef LOGFILE /* { */
	if (lfp) fclose(lfp);
#endif /* } LOGFILE */
	unlink(LOCKFILE);
}

fsize(file)
char *file;
{
	struct stat st;

	if (access(SPOOLFILE,F_OK))
		/* not existing */
		return(0);
	if (stat(SPOOLFILE,&st))
		/* no access */
		return(0);
	/* if (!st.st_size)
		// empty //
		return(0); */
	return(st.st_size);
}

uux()
{
	FILE	*pp;
	FILE	*fp;
#ifdef BACKUP /* { */
	char	*bakfile;
#endif /* } BACKUP */
#if defined(BACKUP) || defined(LOGFILE) /* { */
	long	t = time(0L);
#endif /* } BACKUP || LOGFILE */
#ifdef LOGFILE /* { */
	struct tm *tm;

	tm = localtime(&t);
	if (lfp) fprintf(lfp,"%5d %2d.%2d.%2d %02d:%02d:%02d uux(): ",
			getpid(),
			tm->tm_mday, tm->tm_mon+1, tm->tm_year,
			tm->tm_hour, tm->tm_min, tm->tm_sec);
#endif /* } LOGFILE */
	if (!fsize(SPOOLFILE)) {
#ifdef LOGFILE /* { */
		if (lfp) fprintf(lfp,"not existing, empty or unavailable\n");
#endif /* } LOGFILE */
		return(0);
	}
	if (pp = popen(DELIVER,"w")) {
		if (fp = fopen(SPOOLFILE,"r")) {
#ifdef DEBUG /* { */
			fprintf(stderr,"*** %5d open pipe \"%s\"\n",getpid(),DELIVER);
			fflush(stderr);
#endif /* } DEBUG */
			fprintf(pp,"HELO %s\n",localhost);
			while(fgets(buffer,BUFLEN,fp)) fputs(buffer,pp);
			fprintf(pp,"QUIT\n");
			fclose(fp);
#ifdef BACKUP /* { */
			bakfile = (char*)ckalloc(1,sizeof(SPOOLFILE)+20);
			sprintf(bakfile,"%s.%d.%d",SPOOLFILE,t,getpid());
			rename(SPOOLFILE,bakfile);
# ifdef LOGFILE /* { */
			if (lfp) fprintf(lfp,"%s\n",bakfile);
# endif /* } LOGFILE */
#else /* }{ BACKUP */
			unlink(SPOOLFILE);
# ifdef LOGFILE /* { */
			if (lfp) fprintf(lfp,"processed\n");
# endif /* } LOGFILE */
#endif /* } BACKUP */
			
		}
		pclose(pp);
	}
#ifdef DEBUG /* { */
	else {
		perror("*** popen");
		fprintf(stderr,"*** %5d cannot open pipe \"%s\"\n",getpid(),DELIVER);
		fflush(stderr);
	}
#endif /* } DEBUG */
}

#ifdef DEBUG
debug()
{
	fprintf(stdout,"500 Command unrecognized\n");
	fprintf(stderr,"*** %5d Debug turned on. Keep care of what you're doing!\n",getpid());
	fflush(stderr);
	state |= (RCPT|MAIL);
	if (!sender) sender = strdup("fifi@kaa.han.de");
	if (!recipient) {
		recipient = (charl *)ckalloc(1,sizeof(charl));
		recipient->next = 0;
		recipient->rcpt = strdup("fifi@kaa.han.de");
	}
}
#endif

data()
{
	int	i;
	FILE	*tp,*fp;
	int	term = 0;
	static char tempfile[MAXNAMLEN];
	charl	*rp;
#ifdef AUTOUUX
	long	sz;
#endif
#ifdef RECEIVED /* { */
	int	body = 0;
	int	recv = 0;
#endif /* } RECEIVED */
#ifdef LOGFILE /* { */
	struct tm *tm;
	long size=0, lines=0;
#endif /* } LOGFILE */
#if defined(LOGFILE) || defined(RECEIVED) /* { */
	long	t;

	t = time(0L);
#endif /* } LOGFILE || RECEIVED */
#if defined(LOGFILE) || defined(RECEIVED) /* { */
	tm = localtime(&t);
#endif /* } RECEIVED || LOGFILE */
	if (!(state & MAIL)) {
		fprintf(stdout,"503 Need MAIL command\n");
		return(1);
	}
	if (!(state & RCPT)) {
		fprintf(stdout,"503 Need RCPT (recipient)\n");
		return(1);
	}
	/* open temporary file */
	sprintf(tempfile,"%s/q%d.%d",SPOOLDIR,getpid(),t);
# ifdef DEBUG /* { */
	fprintf(stderr,"*** %5d open tempfile %s\n",getpid(),tempfile);
	fflush(stderr);
# endif /* } DEBUG */
	if (!(tp = fopen(tempfile,"w"))) {
		/* open failed... */
#ifdef DEBUG
		perror("*** open failed");
#endif
		fprintf(stdout,"550 Requested action not taken: tempfile unavailable\n");
		quit();
	}
	if (mode & WRBSMTP) {
		fprintf(tp,"MAIL FROM: <%s>\n",sender);
#ifdef DEBUG /* { */
		fprintf(stderr,"*** %5d MAIL FROM: <%s>\n",getpid(),sender);
		fflush(stderr);
#endif /* } DEBUG */
		rp = recipient;
		while (rp) {
			fprintf(tp,"RCPT TO: <%s>\n",rp->rcpt);
#ifdef DEBUG /* { */
			fprintf(stderr,"*** %5d RCPT TO: <%s>\n",getpid(),rp->rcpt);
			fflush(stderr);
#endif /* } DEBUG */
			rp = rp->next;
		}
		fprintf(tp,"DATA\n");
		fprintf(stdout,"354 Send");
		if (state & BIT7) fprintf(stdout," 7BIT");
		else if (state & BIT8) fprintf(stdout," 8BITMIME");
		fprintf(stdout," message, end with \".\" on a line by itself\n");
		fflush(stdout);
# ifdef DEBUG
		fprintf(stderr,"*** %5d start mail\n",getpid());
		fflush(stderr);
# endif
	}
	while(fgets(buffer,BUFLEN,ifp)) {
		stripcr(buffer);
#ifdef RECEIVED /* { */
		if (!body && !recv && (!*buffer || !strnucmp(buffer,"Received: ",10))) {
			if (!remotehost || !*remotehost) remotehost = localhost;
			if (!actname || !*actname) actname = remotehost;
			fprintf(tp,"%s: from %s (%s)\n\tby %s (%s);\n\t%s",
				RECEIVED, remotehost, actname,
				localhost, vstring, asctime(tm));
			fflush(tp);
			recv++;
		}
		if (!*buffer) body++;
#endif /* } RECEIVED */
		strcat(buffer,"\n");
		if (!fputs(buffer,tp)) {
			/* error occured */
#ifdef DEBUG /* { */
			perror("*** write failed");
#endif /* } DEBUG */
			fprintf(stdout,"552 Requested mail action aborted: exceeded storage allocation\n");
			quit();
		}
		stripcr(buffer);
		if (!strcmp(buffer,".")) {
			term++;
#ifdef DEBUG /* { */
			fprintf(stderr,"*** %5d eof detected\n",getpid());
			fflush(stderr);
#endif /* } DEBUG */
			break;
		}
#ifdef LOGFILE /* { */
		lines++; size += strlen(buffer);
#endif /* } LOGFILE */
	}
	fclose(tp);
#ifdef DEBUG /* { */
	fprintf(stderr,"*** %5d fclose ok\n",getpid(),fp);
	fflush(stderr);
#endif /* } DEBUG */
	if (!term) {
#ifdef DEBUG /* { */
		fprintf(stderr,"*** %5d eof missed\n",getpid());
		fflush(stderr);
#endif /* } DEBUG */
		unlink(tempfile);
		quit();
	}
	/* try to deliver file */
	lock();
#ifdef LOGFILE /* { */
	if (lfp) fprintf(lfp,"%5d %2d.%2d.%2d %02d:%02d:%02d %s %s -> ",
			getpid(),
			tm->tm_mday, tm->tm_mon+1, tm->tm_year,
			tm->tm_hour, tm->tm_min, tm->tm_sec,
			prog,sender);
#endif /* } LOGFILE */
	/* reopen tempfile */
	if (!(tp = fopen(tempfile,"r"))) {
#ifdef DEBUG
		perror("*** reopen failed");
#endif
		fprintf(stdout,"550 Requested action not taken: tempfile reopen failed\n");
		quit();
	}
	if (mode & WRSPOOL) {
# ifdef DEBUG /* { */
		fprintf(stderr,"*** %5d open file %s\n",getpid(),SPOOLFILE);
		fflush(stderr);
# endif /* } DEBUG */
		if (!(fp = fopen(SPOOLFILE,"a+"))) {
# ifdef DEBUG
			perror("*** open failed");
# endif
			fprintf(stdout,"550 Requested action not taken: mailbox unavailable\n");
			quit();
		}
	} else /* if (mode & WRPIPE) */ {
		/* accept data from ifp */
		sprintf(buffer,"%s -f%s ",SENDMAIL,sender);
	}
	while (recipient) {
		if (mode & WRPIPE) {
			strcat(buffer,recipient->rcpt);
			strcat(buffer," ");
		}
#ifdef LOGFILE /* { */
		if (lfp) fprintf(lfp,"%s ",recipient->rcpt);
#endif /* } LOGFILE */
		recipient = recipient->next;
	}
	if (mode & WRPIPE) {
# ifdef DEBUG
		fprintf(stderr,"*** %5d popen(%s)\n",getpid(),buffer);
		fflush(stderr);
# endif
		if (!(fp = popen(buffer,"w"))) {
			fprintf(stderr,"*** %5d popen(%s) failed \n",getpid(),buffer);
			exit(1);
		}
	}
	reset();
	while(fgets(buffer,BUFLEN,tp)) {
		if (!fputs(buffer,fp)) {
			/* error occured */
#ifdef DEBUG /* { */
			perror("*** write failed");
#endif /* } DEBUG */
			fprintf(stdout,"552 Requested mail action aborted: exceeded storage allocation\n");
			quit();
		}
	}
	fclose(tp);
	if (mode & WRSPOOL) fclose(fp);
	   else /* if (mode & WRPIPE) */ pclose(fp);
	unlink(tempfile);
#ifdef DEBUG /* { */
	fprintf(stderr,"*** %5d fclose ok\n",getpid(),fp);
	fflush(stderr);
#endif /* } DEBUG */
#ifdef LOGFILE /* { */
	if (lfp) {
		fprintf(lfp,"(L:%d,S:%d)\n",lines,size);
		fflush(lfp);
	}
#endif /* } LOGFILE */
#if defined(AUTOUUX) /* { */
	if (mode & WRSPOOL) {
		sz = fsize(SPOOLFILE);
# ifdef DEBUG /* { */
		fprintf(stderr,"*** %5d %s: %d (uux: %d)\n",getpid(),SPOOLFILE,sz,SPOOLSIZE);
		fflush(stderr);
# endif /* } DEBUG */
		if (sz > SPOOLSIZE) uux();
	}
#endif /* } AUTOUUX */
	unlock();
	fprintf(stdout,"250 Ok\n");
	return(0);
}

reset()
{
	sender = 0;
	recipient = 0;
	state &= ~(BIT7|BIT8|MAIL|RCPT);
}

rset()
{
	reset();
	fprintf(stdout,"250 Reset state\n");
	return(0);
}

rcpt()
{
	char	*cp,*sp;
	charl	*cl;

	if (!(cp = strchr(buffer,':'))) {
		fprintf(stdout,"501 Syntax error\n");
		return(1);
	}
	cp = skip20(cp+1);
	if (!cp || !*cp) return(0);
	while (*cp == '<') {
		cp++;
		if (sp = strrchr(cp,'>')) *sp = 0;
	}
	state |= RCPT;
	cl = (charl*)ckalloc(1,sizeof(charl));
	cl->next = recipient;
	cl->rcpt = strdup(cp);
	recipient = cl;
	fprintf(stdout,"250 %s... Recipient ok\n",recipient->rcpt);
	return(0);
}

mail()
{
	char	*cp;
	char	*sp;
#if defined(RFC1653) /* { */
	long	s = 0;
#endif /* } */

	if (state & MAIL) {
		fprintf(stdout,"554 Nested MAIL command\n");
		return(1);
	}
	if (!(cp = strchr(buffer,':'))) {
		fprintf(stdout,"501 Syntax error\n");
		return(1);
	}
	state |= MAIL;
	cp = skip20(cp+1);
	if (!cp || !*cp) cp = "root";
#if defined(RFC1653) || defined(RFC1652) /* { */
	if (state & EHLO) {
		/* look for size parameter */
		if (sp = skipch(cp,' ')) *sp++ = 0;
		while (sp && *sp && (sp = skip20(sp))) {
# if defined(RFC1652) /* { */
			if (!strnucmp(sp,"body=",5)) {
				sp += 5;
#ifdef DEBUG
				fprintf(stderr,"*** %5d got body=%s\n",getpid(),sp);
#endif
				if (!struncmp(sp,"7bit",4)) {
					state |= BIT7;	
				} else if (!struncmp(sp,"8bitmime",8)) {
					state |= BIT8;
				} else {
					fprintf(stdout,"501 Syntax error in parameter body\n");
					reset();
					return(1);
				}
				sp = skip2space(sp);
				continue;
			}
# endif /* } RFC1652 */
# if defined(RFC1653) /* { */
			if (!strnucmp(sp,"size=",5)) {
				s = atol(sp+5);
#ifdef DEBUG
				fprintf(stderr,"*** %5d got size=%d\n",getpid(),s);
#endif
				if (((s+1023)/1024) > sysspace) {
					/* out of memory */
					fprintf(stdout,"452 Message too long\n");
					reset();
					return(1);
				}
				sp = skip2space(sp);
				continue;
			}
# endif /* } RFC1653 */
			fprintf(stdout,"501 Syntax error in arguments: %s\n",sp);
			reset();
			return(1);
		}
	}
#endif /* } RFC1653 || RFC1652 */
	while (*cp == '<') {
		cp++;
		if (sp = strrchr(cp,'>')) *sp = 0;
	}
	if (!*cp) {
		sender = (char*)ckalloc(1,(strlen("Mailer-Deamon")+strlen(remotehost)+5));
		sprintf(sender,"%s!%s",remotehost,"Mailer-Deamon");
	} else {
		sender = strdup(cp);
	}
	fprintf(stdout,"250 %s... Sender ok\n",sender);
	return(0);
}

vrfy()
{
	char	*cp;
	cp = skip20(buffer+sizeof("HELO"));

	fprintf(stdout,"250 %s (unchecked)\n",cp);
	return(0);
}

#if defined(RFC1652) || defined(RFC1653) /* { */
ehlo()
{
	helo();
	state = EHLO;
#ifdef MIME
# if defined(RFC1652) || defined(RFC1653) /* { */
	fprintf(stdout,"250-MIME\n");
# else
	fprintf(stdout,"250 MIME\n");
# endif /* } */
#endif
# ifdef RFC1653 /* { */
#  ifdef RFC1652 /* { */
	fprintf(stdout,"250-SIZE %d\n",RFC1653*1024);
#  else
	fprintf(stdout,"250 SIZE %d\n",RFC1653*1024);
#  endif
# endif /* } RFC1653 */
# ifdef RFC1652 /* { */
	fprintf(stdout,"250 8BITMIME\n");
# endif /* } RFC1653 */
	return(0);
}
#endif /* } RFC1652 | RFC1653 */

helo()
{
	char	*cp;

	state = HELO;
	cp = skip20(buffer+sizeof("HELO"));
	actname = strdup(cp);
	if (!remotehost) remotehost = actname;
#if defined(MIME) || defined(RFC1652) || defined(RFC1653) /* { */
	if ((*buffer == 'E') || (*buffer == 'e'))
		fprintf(stdout,"250-");
	else
#endif /* } */
	fprintf(stdout,"250 ");
	fprintf(stdout,"%s Hi %s, ",localhost,remotehost);
	if (!strucmp(remotehost,cp)) fprintf(stdout,"nice to meet you\n");
		else fprintf(stdout,"why do you call yourself %s?\n",cp);
	return(0);
}

noop()
{
	fprintf(stdout,"200 OK\n");
	return(0);
}

quit()
{
	fprintf(stdout,"221 %s closing connection\n",localhost);
#ifdef DEBUG
	fprintf(stderr,"*** %5d ready...\n",getpid());
	fflush(stderr);
#endif
	fflush(stdout);
	exit(0);
}

help()
{
	int	i;

	fprintf(stdout,"214-Commands:\n");
	for(i=0; ftab[i].cmd; i++) {
		if (!(i%5)) fprintf(stdout,"214-");
		fprintf(stdout,"    %s",ftab[i].cmd);
		if ((i%5) == 4) fprintf(stdout,"\n");
	}
	if ((i%5) != 4) fprintf(stdout,"\n");
	fprintf(stdout,"214 End of help info\n");
	return(0);
}

main(argc,argv)
int	argc;
char	**argv;
{
	char	*cp;
	struct tm	*tm;
	long	t;
	int	i;
	FILE	*fp;

	int fromlen;
	struct sockaddr_in from;
	struct hostent *hp;
#ifdef STATFS
	struct statfs sfs;
#endif
#ifdef USTAT
	struct	stat st;
	struct	ustat ust;
#endif


	prog = basename(argv[0]);
#if defined(DEBUG) && defined(DEBUGDEV)
	if (i = open(DEBUGDEV,O_CREAT|O_WRONLY|O_APPEND,0600)) {
		close(2);
		dup(i);
		close(i);
	}
#endif

	gethostname(buffer,BUFLEN);
	localhost = strdup(buffer);
	t = time(0L);
	tm = localtime(&t);

	sprintf(vstring,"%s %s",prog,VERSION);
#ifdef MIME
	strcat(vstring,"m");
#endif
#ifdef RFC1652
	strcat(vstring,"M");
#endif
#ifdef RFC1653
	strcat(vstring,"S");
#endif

# ifdef DEBUG
	fprintf(stderr,"*** %5d %s (%s) ready\n",getpid(),argv[0],vstring);
	fflush(stderr);
# endif

#ifdef CALLUUX /* { */
	if (!strcmp(prog,CALLUUX)) {
		if (!fsize(SPOOLFILE)) exit(0);
		lock();
		uux();
		unlock();
		exit(0);
	}
#endif /* } CALLUUX */
	if (!strcmp(prog,BSMTPD)) {
		mode = STATE_BSMTPD;
		fromlen = sizeof (from);
		if (getpeername(0, &from, &fromlen) < 0) {
			fprintf(stdout,"522 Can't get peer name of remote host\n");
			quit();
		}
		from.sin_port = ntohs((u_short)from.sin_port);
		if (!(hp = gethostbyaddr((char*)&from.sin_addr, sizeof (struct in_addr), from.sin_family)))
			remotehost = (char*)inet_ntoa(from.sin_addr);
		else
			remotehost = hp->h_name;
# ifdef DEBUG
		fprintf(stderr,"*** %5d remote host: %s\n",getpid(),remotehost);
		fflush(stderr);
# endif
	}
	if (!strcmp(prog,RBSMTP)) {
		mode = STATE_RBSMTP;
		close(1);
		dup(2);
	}

	if (mode & CHECKSIZE) {
		/*
		 * check for existence of files and directories and
		 * for MINFREE kB free space
		 */
# ifdef DEBUG
		fprintf(stderr,"*** %5d Check spool file %s\n",getpid(),SPOOLFILE);
		fflush(stderr);
# endif
		if (!(fp = fopen(SPOOLFILE,"a+"))) {
			fprintf(stdout,"550 Requested action not taken: mailbox unavailable\n");
			quit();
		}
		fclose(fp);
#if defined(STATFS) || defined(USTAT) /* { */
# ifdef STATFS /* { */
		if (statfs(SPOOLFILE,&sfs)) {
			fprintf(stdout,"551 Requested action aborted: error in processing: statfs\n");
			quit();
		} else {
			sysspace = (sfs.f_bavail * sfs.f_bsize)/1024;
#  ifdef DEBUG
			fprintf(stderr,"*** %5d System storage: %d kB left (free=%d, avail=%d, unit=%d)\n",getpid(),sysspace,sfs.f_bfree,sfs.f_bavail,sfs.f_bsize);
			fflush(stderr);
#  endif
			if (sysspace < MINFREE) {
				fprintf(stdout,"452 Requested action not taken: insufficient system storage\n");
				quit();
			}
		}
# else /* STATFS } { USTAT */
		if (stat(SPOOLFILE,&st)) {
			fprintf(stdout,"551 Requested action aborted: error in processing: stat\n");
			quit();
		}
		if (ustat(st.st_dev,&ust)) {
			fprintf(stdout,"551 Requested action aborted: error in processing: ustat\n");
			quit();
		} else {
			sysspace = (ust.f_tfree * BBSIZE)/1024;
#  ifdef DEBUG
			fprintf(stderr,"*** %5d System storage: %d kB left (free=%d, unit=%d)\n",getpid(),sysspace,ust.f_tfree,BBSIZE);
			fflush(stderr);
#  endif
			if (sysspace < MINFREE) {
				fprintf(stdout,"452 Requested action not taken: insufficient system storage\n");
				quit();
			}
		}
# endif /* } USTAT */
#else /*  USTAT | STATFS } { */
# ifdef DEBUG
		fprintf(stderr,"*** %5d no system storage size information\n",getpid());
# endif
		sysspace = 0;
#endif /* } STATFS || USTAT */
	}
	/*
	 * Greetings...
	 */
	fprintf(stdout,"220 %s %s ready at %s",
		localhost,vstring,asctime(tm));
	fflush(stdout);

	if (mode & RDSTDINC) {
		if (!(ifp = depress(stdin))) {
			fprintf(stderr,"*** %5d error depress\n",getpid());
			exit(1);
		}
	}

	while(fgets(buffer,BUFLEN,ifp)) {
		strip(buffer);
#ifdef DEBUG
		fprintf(stderr,">>> %5d \"%s\"\n",getpid(),buffer);
		fflush(stderr);
#endif
		for(i=0; ftab[i].cmd && strnucmp(ftab[i].cmd,buffer,strlen(ftab[i].cmd)); i++);
		if (ftab[i].cmd) {
			ftab[i].func();	
		} else {
			fprintf(stdout,"500 Command unrecognized\n");
			if (mode & RDSTDIN) exit(1);
		}
		fflush(stdout);
	}
	fclose(ifp);
	return(0);
}
	
